//----------------------------------------------------------------------------
/** @file GtpEngine.cpp
    @see GtpEngine.h
*/
//----------------------------------------------------------------------------

#include "GtpEngine.h"

// Only include standard C++ headers in GtpEngine.cpp
#include <iomanip>
#include <cassert>
#include <cctype>
#include <fstream>

#if GTPENGINE_PONDER
#include <boost/thread/barrier.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/thread.hpp>
#endif

using namespace std;

#if GTPENGINE_PONDER
using boost::barrier;
using boost::condition;
using boost::mutex;
using boost::thread;
#endif

//----------------------------------------------------------------------------

#if GTPENGINE_PONDER

namespace {

/** Function object used by GtpEngine::MainLoop().
    This function runs in a different thread and calls GtpEngine::Ponder()
    while the engine is waiting for the next command.
    @see GtpEngine::Ponder()
*/
class PonderFunction
{
public:
    PonderFunction(GtpEngine& engine, barrier& ponderThreadReady,
                   mutex& startPonderMutex, mutex& ponderFinishedMutex,
                   condition& startPonder, condition& ponderFinished);

    void operator()();

private:
    GtpEngine& m_engine;

    barrier& m_ponderThreadReady;

    mutex& m_startPonderMutex;

    mutex& m_ponderFinishedMutex;

    condition& m_startPonder;

    condition& m_ponderFinished;
};


PonderFunction::PonderFunction(GtpEngine& engine, barrier& ponderThreadReady,
                               mutex& startPonderMutex,
                               mutex& ponderFinishedMutex,
                               condition& startPonder,
                               condition& ponderFinished)
    : m_engine(engine),
      m_ponderThreadReady(ponderThreadReady),
      m_startPonderMutex(startPonderMutex),
      m_ponderFinishedMutex(ponderFinishedMutex),
      m_startPonder(startPonder),
      m_ponderFinished(ponderFinished)
{
}

void PonderFunction::operator()()
{
    mutex::scoped_lock lock(m_startPonderMutex);
    m_ponderThreadReady.wait();
    while (true)
    {
        m_startPonder.wait(lock);
        if (m_engine.IsQuitSet())
            return;
        m_engine.Ponder();
        {
            mutex::scoped_lock lock(m_ponderFinishedMutex);
            m_ponderFinished.notify_all();
        }
    }
}

} // namespace

#endif // GTPENGINE_PONDER

//----------------------------------------------------------------------------

namespace {

void Trim(string& str);

/** Check, if line contains a command.
    @param line The line to check.
    @return True, if command does not contain only whitespaces and is not a
    comment line.
*/
bool IsCommandLine(const string& line)
{
    string trimmedLine = line;
    Trim(trimmedLine);
    return (! trimmedLine.empty() && trimmedLine[0] != '#');
}

/** Replace empty lines in a multi-line string by lines containing a single
    space.
    @param text The input string.
    @return The input string with all occurrences of "\n\n" replaced by
    "\n \n".
*/
string ReplaceEmptyLines(const string& text)
{
    if (text.find("\n\n") == string::npos)
        return text;
    istringstream in(text);
    ostringstream result;
    bool lastWasNewLine = false;
    char c;
    while (in.get(c))
    {
        bool isNewLine = (c == '\n');
        if (isNewLine && lastWasNewLine)
            result.put(' ');
        result.put(c);
        lastWasNewLine = isNewLine;
    }
    return result.str();
}

/** Split line into arguments.
    Arguments are words separated by whitespaces.
    Arguments with whitespaces can be quoted with quotation marks ('"').
    Characters can be escaped with a backslash ('\').
    @param line The line to split.
    @return A vector containing the arguments.
*/
vector<string> SplitLine(const string& line)
{
    vector<string> result;
    bool escape = false;
    bool inString = false;
    ostringstream element;
    for (size_t i = 0; i < line.size(); ++i)
    {
        char c = line[i];
        if (c == '"' && ! escape)
        {
            if (inString)
            {
                result.push_back(element.str());
                element.str("");
            }
            inString = ! inString;
        }
        else if (isspace(c) && ! inString)
        {
            if (! element.str().empty())
            {
                result.push_back(element.str());
                element.str("");
            }
        }
        else
            element << c;
        escape = (c == '\\' && ! escape);
    }
    if (! element.str().empty())
        result.push_back(element.str());
    return result;
}

/** Remove leading and trailing whitespaces from a string.
    Whitespaces are tab, carriage return and space.
    @param str The input string.
*/
void Trim(string& str)
{
    char const* whiteSpace = " \t\r";
    size_t pos = str.find_first_not_of(whiteSpace);
    str.erase(0, pos);
    pos = str.find_last_not_of(whiteSpace); 
    str.erase(pos + 1);
}

} // namespace

//----------------------------------------------------------------------------

GtpFailure::GtpFailure()
{
}

GtpFailure::GtpFailure(const GtpFailure& failure)
{
    m_response << failure.Response();
    m_response.copyfmt(failure.m_response);
}
    
GtpFailure::GtpFailure(const std::string& response)
{
    m_response << response;
}

GtpFailure::~GtpFailure() throw()
{
}

//----------------------------------------------------------------------------

ostringstream GtpCommand::s_dummy;

const std::string& GtpCommand::Arg(std::size_t number) const
{
    if (number >= NuArg())
        throw GtpFailure() << "missing argument " << (number + 1);
    return m_arguments[number];
}

std::string GtpCommand::ArgToLower(std::size_t number) const
{
    string value = Arg(number);
    for (size_t i = 0; i < value.length(); ++i)
        value[i] = tolower(value[i]);
    return value;
}

bool GtpCommand::BoolArg(std::size_t number) const
{
    string value = Arg(number);
    if (value == "1")
        return true;
    if (value == "0")
        return false;
    throw GtpFailure() <<  "argument " << (number + 1) << " must be 0 or 1";
}

void GtpCommand::CheckNuArg(std::size_t number) const
{
    if (NuArg() == number)
        return;
    throw GtpFailure() << "command needs " << number << " argument(s)";
}

void GtpCommand::CheckNuArgLessEqual(std::size_t number) const
{
    if (NuArg() <= number)
        return;
    throw GtpFailure() << "command needs at most " << number
                       << " argument(s)";
}

double GtpCommand::FloatArg(std::size_t number) const
{
    istringstream in(Arg(number));
    double result;
    in >> result;
    if (! in)
        throw GtpFailure() << "argument " << (number + 1)
                           << " must be a float";
    return result;
}

int GtpCommand::IntArg(std::size_t number) const
{
    istringstream in(Arg(number));
    int result;
    in >> result;
    if (! in)
        throw GtpFailure() << "argument " << (number + 1)
                           << " must be a number";
    return result;
}

int GtpCommand::IntArg(std::size_t number, int min) const
{
    int result = IntArg(number);
    if (result < min)
        throw GtpFailure() << "argument " << (number + 1)
                           << " must be >= " << min;
    return result;
}

int GtpCommand::IntArg(std::size_t number, int min, int max) const
{
    int result = IntArg(number);
    if (result < min)
        throw GtpFailure() << "argument " << (number + 1)
                           << " must be >= " << min;
    if (result > max)
        throw GtpFailure() << "argument " << (number + 1)
                           << " must be <= " << max;
    return result;
}

void GtpCommand::Init(const std::string& line)
{
    m_line = line;
    Trim(m_line);
    m_arguments = SplitLine(m_line);
    assert(m_arguments.size() > 0);
    ParseCommandId();
    assert(m_arguments.size() > 0);
    m_name = m_arguments[0];
    m_arguments.erase(m_arguments.begin());
    m_argLine = line.substr(m_line.find(m_name) + m_name.size());
    Trim(m_argLine);
    m_response.str("");
    m_response.copyfmt(s_dummy);
}

void GtpCommand::ParseCommandId()
{
    m_id = "";
    if (m_arguments.size() < 2)
        return;
    istringstream in(m_arguments[0]);
    int id;
    in >> id;
    if (in)
    {
        m_id = m_arguments[0];
        m_arguments.erase(m_arguments.begin());
    }
}

void GtpCommand::SetResponse(const std::string& response)
{
    m_response.str(response);
}

void GtpCommand::SetResponseBool(bool value)
{
    m_response.str(value ? "true" : "false");
}

//----------------------------------------------------------------------------

GtpCallbackBase::~GtpCallbackBase() throw()
{
}

//----------------------------------------------------------------------------

GtpEngine::GtpEngine(std::istream& in, std::ostream& out)
    : m_in(in),
      m_out(out)
      
{
    Register("known_command", &GtpEngine::CmdKnownCommand);
    Register("list_commands", &GtpEngine::CmdListCommands);
    Register("name", &GtpEngine::CmdName);
    Register("protocol_version", &GtpEngine::CmdProtocolVersion);
    Register("quit", &GtpEngine::CmdQuit);
    Register("version", &GtpEngine::CmdVersion);
}

GtpEngine::~GtpEngine()
{
    for (CallbackMap::iterator callback = m_callbacks.begin();
         callback != m_callbacks.end(); ++callback)
    {
        delete callback->second;
#ifndef NDEBUG
        callback->second = 0;
#endif
    }
}

void GtpEngine::BeforeHandleCommand()
{
    // Default implementation does nothing
}

/** Return <code>true</code> if command is known, <code>false</code>
    otherwise.
*/
void GtpEngine::CmdKnownCommand(GtpCommand& cmd)
{
    cmd.CheckNuArg(1);
    cmd.SetResponseBool(IsRegistered(cmd.Arg(0)));
}

/** List all known commands. */
void GtpEngine::CmdListCommands(GtpCommand& cmd)
{
    cmd.CheckArgNone();
    for (CallbackMap::const_iterator i = m_callbacks.begin();
         i != m_callbacks.end(); ++i)
        cmd << i->first << '\n';
}

/** Return name. */
void GtpEngine::CmdName(GtpCommand& cmd)
{
    cmd.CheckArgNone();
    cmd << "Unknown";
}

/** Return protocol version. */
void GtpEngine::CmdProtocolVersion(GtpCommand& cmd)
{
    cmd.CheckArgNone();
    cmd << "2";
}

/** Quit command loop. */
void GtpEngine::CmdQuit(GtpCommand& cmd)
{
    cmd.CheckArgNone();
    SetQuit();
}

/** Return empty version string.
    The GTP standard says to return empty string, if no meaningful reponse
    is available.
*/
void GtpEngine::CmdVersion(GtpCommand& cmd)
{
    cmd.CheckArgNone();
}

void GtpEngine::ExecuteCommand(const string& cmdline, std::ostream& log)
{
    if (! IsCommandLine(cmdline))
        throw GtpFailure() << "Bad command: " << cmdline;
    GtpCommand cmd;
    cmd.Init(cmdline);
    log << cmd.Line() << '\n';
    bool status = HandleCommand(cmd, log);
    if (! status)
        throw GtpFailure() << "Executing " << cmd.Line() << " failed: "
                           << cmd.Response();
}

void GtpEngine::ExecuteFile(const string& name, std::ostream& log)
{
    ifstream in(name.c_str());
    if (! in)
        throw GtpFailure() << "Cannot read " << name;
    string line;
    GtpCommand cmd;
    while (getline(in, line))
    {
        if (! IsCommandLine(line))
            continue;
        cmd.Init(line);
        log << cmd.Line() << '\n';
        bool status = HandleCommand(cmd, log);
        if (! status)
            throw GtpFailure() << "Executing " << cmd.Line() << " failed: "
                               << cmd.Response();
    }
}

bool GtpEngine::HandleCommand(GtpCommand& cmd, ostream& out)
{
    BeforeHandleCommand();
    bool status = true;
    string response;
    try
    {
        CallbackMap::const_iterator pos = m_callbacks.find(cmd.Name());
        if (pos == m_callbacks.end())
        {
            status = false;
            response = "unknown command: " + cmd.Name();
        }
        else
        {
            GtpCallbackBase* callback = pos->second;
            (*callback)(cmd);
            response = cmd.Response();
        }
    }
    catch (const GtpFailure& failure)
    {
        status = false;
        response = failure.Response();
    }
    response = ReplaceEmptyLines(response);
    out << (status ? '=' : '?');
    out << cmd.ID();
    out << ' ' << response;
    size_t size = response.size();
    if (size == 0 || response[size - 1] != '\n')
        out << '\n';
    out << endl;
    return status;
}

bool GtpEngine::IsRegistered(const std::string& command) const
{
    return (m_callbacks.find(command) != m_callbacks.end());
}

void GtpEngine::MainLoop()
{
    m_quit = false;
#if GTPENGINE_PONDER
    mutex startPonderMutex;
    mutex ponderFinishedMutex;
    condition startPonder;
    condition ponderFinished;
    barrier ponderThreadReady(2);
    PonderFunction ponderFunction(*this, ponderThreadReady,
                                  startPonderMutex, ponderFinishedMutex,
                                  startPonder, ponderFinished);
    boost::thread ponderThread(ponderFunction);
    ponderThreadReady.wait();
    mutex::scoped_lock lock(ponderFinishedMutex);
#endif
    vector<string> arguments;
    string line;
    GtpCommand cmd;
    while (true)
    {
#if GTPENGINE_PONDER
        InitPonder();
        {
            mutex::scoped_lock lock(startPonderMutex);
            startPonder.notify_all();
        }
#endif
        while (getline(m_in, line) && ! IsCommandLine(line));
#if GTPENGINE_PONDER
        StopPonder();
        ponderFinished.wait(lock);
#endif
        if (m_in)
        {
            Trim(line);
            cmd.Init(line);
            HandleCommand(cmd, m_out);
        }
        else
            SetQuit();
        if (m_quit)
        {
#if GTPENGINE_PONDER
            {
                mutex::scoped_lock lock(startPonderMutex);
                startPonder.notify_all();
            }
            ponderThread.join();
#endif
            break;
        }
    }
}

void GtpEngine::Register(const std::string& command,
                         GtpCallbackBase* callback)
{
    CallbackMap::iterator pos = m_callbacks.find(command);
    if (pos != m_callbacks.end())
    {
        delete pos->second;
#ifndef NDEBUG
        pos->second = 0;
#endif
        m_callbacks.erase(pos);
    }
    m_callbacks.insert(make_pair(command, callback));
}

void GtpEngine::Register(const std::string& command,
                         GtpCallback<GtpEngine>::Method method)
{
    Register(command, new GtpCallback<GtpEngine>(this, method));
}

void GtpEngine::SetQuit()
{
    m_quit = true;
}

bool GtpEngine::IsQuitSet() const
{
    return m_quit;
}

#if GTPENGINE_PONDER

void GtpEngine::Ponder()
{
}

void GtpEngine::StopPonder()
{
}

void GtpEngine::InitPonder()
{
}

#endif // GTPENGINE_PONDER

//----------------------------------------------------------------------------
